%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%       {log} IMPLEMENTATION AND AUTOMATIC VERIFICATION        %
%                            OF THE                            %
%                 BELL-LAPADULA SECURITY MODEL                 %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%                   THE STATE OF THE SYSTEM                    %
%                                                              %
% SState is the tuple of state variables of the following form %
%                                                              %
%      SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]]        %
%                                                              %
% Each component is as follows:                                % %                                                              %
% Br: is a binary relation such that [S,[O,L]] in Br iff       %
% subject S has opened object O in read mode and O's access    %
% class is L                                                   %
%                                                              %
% Bw: is a binary relation such that [S,[O,L]] in Bw iff       %
% subject S has opened object O in write mode and O's access   %
% class is L                                                   %
%                                                              %
% Fo: is a partial function from objects to access classes     %
% representing the access class of each object                 %
%                                                              %
% Fs: is a partial function from subjects to access classes    %
% representing the access class of each subject                %
%                                                              %
% M: is a binary relation such that [O,[S,A]] in M iff subject %
% S has the right to access object O in mode A. It represents  %
% the access matrix.                                           %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%                         PROPERTIES                           %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% the dominates relation

dominates([L1,C1],[L2,C2]) :-
  subset(C1,C2) & L1 =< L2.

n_dominates([L1,C1],[L2,C2]) :-
  neg(subset(C1,C2) & L1 =< L2).

% security condition or simple security

seccond(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  foreach([Si,[Oi,Sco]] in Br,
    let([Scs],applyTo(Fs,Si,Scs),dominates(Sco,Scs))
  ).

n_seccond(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(
    foreach([Si,[Oi,Sco]] in Br,
      let([Scs],applyTo(Fs,Si,Scs),dominates(Sco,Scs))
    )
  ).

% *-property

starprop(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  foreach([[S1,[O1,Sco1]] in Br,[S2,[O2,Sco2]] in Bw],
    S1 = S2 implies dominates(Sco1,Sco2)
  ).

n_starprop(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(
    foreach([[S1,[O1,Sco1]] in Br,[S2,[O2,Sco2]] in Bw],
      S1 = S2 implies dominates(Sco1,Sco2)
    )
  ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%                         OPERATIONS                           %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
% SState represents the before state                           %
%                                                              %
% SState_ represents the after state                           %
%                                                              %
% Out represents the error code returned by the operation      %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% getRead: subject S requests read access to object O

getRead(SState,S,O,SState_,Out) :-
  getReadOk(SState,S,O,SState_,Out)
  or
  getReadE1(SState,S,O,SState_,Out)
  or
  getReadE2(SState,S,O,SState_,Out)
  or
  getReadE3(SState,S,O,SState_,Out)
  or
  getReadE4(SState,S,O,SState_,Out)
  or
  getReadE5(SState,S,O,SState_,Out)
  or
  getReadE6(SState,S,O,SState_,Out).

% where

getReadOk(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,read]] in M &
  [S,[O,Sco]] nin Br &
  applyTo(Fo,O,Sco) &
  applyTo(Fs,S,Scs) &
  dominates(Sco,Scs) &
  foreach([Si,[Oi,Scoi]] in Bw, Si = S implies dominates(Sco,Scoi)) &
  Br_ = {[S,[O,Sco]]/Br} &
  SState_ = [[br,Br_],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  Out = yes.
getReadE1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,read]] in M) & 
  SState_ = SState & 
  Out = no.
getReadE2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  neg([S,[O,Sco]] nin Br) &
  SState_ = SState & 
  Out = no.
getReadE3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  comp({[O,O]},Fo,{}) & 
  SState_ = SState & 
  Out = no.
getReadE4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  comp({[S,S]},Fs,{}) & 
  SState_ = SState & 
  Out = no.
getReadE5(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  applyTo(Fs,S,Scs) &
  neg(dominates(Sco,Scs)) &
  SState_ = SState & 
  Out = no.
getReadE6(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  neg(foreach([Si,[Oi,Scoi]] in Bw, Si = S implies dominates(Sco,Scoi))) &
  SState_ = SState & 
  Out = no.

% getWrite: subject S requests write access to object O

getWrite(SState,S,O,SState_,Out) :-
  getWriteOk(SState,S,O,SState_,Out)
  or
  getWriteE1(SState,S,O,SState_,Out)
  or
  getWriteE2(SState,S,O,SState_,Out)
  or
  getWriteE3(SState,S,O,SState_,Out)
  or
  getWriteE4(SState,S,O,SState_,Out).

% where

getWriteOk(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,write]] in M &
  [S,[O,Sco]] nin Bw &
  applyTo(Fo,O,Sco) &
  foreach([Si,[Oi,Scoi]] in Br, Si neq S or dominates(Scoi,Sco)) &
  Bw_ = {[S,[O,Sco]]/Bw} &
  SState_ = [[br,Br],[bw,Bw_],[fo,Fo],[fs,Fs],[m,M]] &
  Out = yes.
getWriteE1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,write]] in M) & 
  SState_ = SState & 
  Out = no.
getWriteE2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  neg([S,[O,Sco]] nin Bw) &
  SState_ = SState & 
  Out = no.
getWriteE3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  comp({[O,O]},Fo,{}) & 
  SState_ = SState & 
  Out = no.
getWriteE4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  neg(foreach([Si,[Oi,Scoi]] in Br, Si neq S or dominates(Scoi,Sco))) &
  SState_ = SState & 
  Out = no.

% releaseRead: subject S does not access object O in read mode anymore

releaseRead(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  diff(Br,{[S,[O,Sco]]},Br_) & 
  SState_ = [[br,Br_],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  Out = yes.

% releaseWrite: subject S does not access object O in write mode anymore

releaseWrite(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  diff(Bw,{[S,[O,Sco]]},Bw_) & 
  SState_ = [[br,Br],[bw,Bw_],[fo,Fo],[fs,Fs],[m,M]] &
  Out = yes.

% changeFo: the new security class of object O is [L,C]

changeFo(SState,O,L,C,SState_,Out) :-
  changeFoOk(SState,O,L,C,SState_,Out)
  or
  changeFoE1(SState,O,L,C,SState_,Out).

% where

changeFoOk(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  L >= 0 &
  set(C) &
  (Fo = {[O,[L,C]]/Fo1} &
   comp(Br,{[[O,[L,C]],[O,[L,C]]]},{}) &
   comp(Bw,{[[O,[L,C]],[O,[L,C]]]},{}) &
   Fo_ = {[O,[L,C]] / Fo1} 
   or 
   comp({[O,O]},Fo,{}) & 
   Fo_ = {[O,[L,C]] / Fo}
  ) &
  SState_ = [[br,Br],[bw,Bw],[fo,Fo_],[fs,Fs],[m,M]] &
  Out = yes.
changeFoE1(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,[L,C]) &
  neg(
    comp(Br,{[[O,[L,C]],[O,[L,C]]]},{}) & comp(Bw,{[[O,[L,C]],[O,[L,C]]]},{})
  ) &
  SState_ = SState &
  Out = no.

% giveRW: subject S grants to subject Si read or write access to object O

giveRW(SState,S,Si,O,X,SState_,Out) :-
  giveRWOk(SState,S,Si,O,X,SState_,Out)
  or
  giveRWE1(SState,S,Si,O,X,SState_,Out)
  or
  giveRWE2(SState,S,Si,O,X,SState_,Out)
  or
  giveRWE3(SState,S,Si,O,X,SState_,Out).

% where

giveRWOk(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  X in {read,write} &
  [O,[S,X]] in M &
  [O,[S,ctrl]] in M &
  M_ = {[O,[Si,X]]/M} & [O,[Si,X]] nin M &
  SState_ = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M_]] &
  Out = yes.
giveRWE1(SState,S,Si,O,X,SState_,Out) :-
  neg(X in {read,write}) &
  SState_ = SState &
  Out = no.
giveRWE2(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  X in {read,write} &
  neg([O,[S,X]] in M) &
  SState_ = SState &
  Out = no.
giveRWE3(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,ctrl]] in M) &
  SState_ = SState &
  Out = no.

% rescindRead: subject S revokes the read access subject Si had on object O

rescindRead(SState,S,Si,O,SState_,Out) :-
  rescindReadOk(SState,S,Si,O,SState_,Out)
  or
  rescindReadE1(SState,S,Si,O,SState_,Out)
  or
  rescindReadE2(SState,S,Si,O,SState_,Out).

% where

rescindReadOk(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,read]] in M &
  [O,[S,ctrl]] in M &
  diff(M,{[O,[Si,read]]},M_) & 
  applyTo(Fo,O,Sco) &
  diff(Br,{[Si,[O,Sco]]},Br_) & 
  SState_ = [[br,Br_],[bw,Bw],[fo,Fo],[fs,Fs],[m,M_]] &
  Out = yes.
rescindReadE1(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,read]] in M) &
  SState_ = SState &
  Out = no.
rescindReadE2(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,ctrl]] in M) &
  SState_ = SState &
  Out = no.

% rescindWrite: subject S revokes the write access subject Si had on object O

rescindWrite(SState,S,Si,O,SState_,Out) :-
  rescindWriteOk(SState,S,Si,O,SState_,Out)
  or
  rescindWriteE1(SState,S,Si,O,SState_,Out)
  or
  rescindWriteE2(SState,S,Si,O,SState_,Out).

% where

rescindWriteOk(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,write]] in M &
  [O,[S,ctrl]] in M &
  diff(M,{[O,[Si,write]]},M_)  &
  applyTo(Fo,O,Sco) &
  diff(Bw,{[Si,[O,Sco]]},Bw_) & 
  SState_ = [[br,Br],[bw,Bw_],[fo,Fo],[fs,Fs],[m,M_]] &
  Out = yes.
rescindWriteE1(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,write]] nin M &
  SState_ = SState &
  Out = no.
rescindWriteE2(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  [O,[S,ctrl]] nin M &
  SState_ = SState &
  Out = no.

% createObject: subject S creates object O

createObject(SState,S,O,SState_,Out) :-
  createObjectOk(SState,S,O,SState_,Out)
  or
  createObjectE1(SState,S,O,SState_,Out)
  or
  createObjectE2(SState,S,O,SState_,Out).

% where

createObjectOk(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  ncomp({[S,S]},Fs,{}) &
  comp({[O,O]},M,{}) &
  M_ = {[O,[S,read]],[O,[S,write]],[O,[S,ctrl]]/M} & 
  SState_ = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M_]] &
  Out = yes.
createObjectE1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(ncomp({[S,S]},Fs,{})) &
  SState_ = SState &
  Out = no.
createObjectE2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(comp({[O,O]},M,{})) &
  SState_ = SState &
  Out = no.

% deleteObject: subject S deletes object O from the system

deleteObject(SState,S,O,SState_,Out) :-
  deleteObjectOk(SState,S,O,SState_,Out)
  or
  deleteObjectE1(SState,S,O,SState_,Out)
  or
  deleteObjectE2(SState,S,O,SState_,Out)
  or
  deleteObjectE3(SState,S,O,SState_,Out).

% where

deleteObjectOk(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  ncomp({[S,S]},Fs,{}) &
  [O,[S,ctrl]] in M &
  applyTo(Fo,O,Sco) &
  comp(Br,{[[O,Sco],[O,Sco]]},{}) &
  comp(Bw,{[[O,Sco],[O,Sco]]},{}) &
  dares({O},M,M_) & 
  SState_ = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M_]] &
  Out = yes.
deleteObjectE1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(ncomp({[S,S]},Fs,{})) &
  SState_ = SState &
  Out = no.
deleteObjectE2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg([O,[S,ctrl]] in M) &
  SState_ = SState &
  Out = no.
deleteObjectE3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  applyTo(Fo,O,Sco) &
  neg(comp(Br,{[[O,Sco],[O,Sco]]},{}) & comp(Bw,{[[O,Sco],[O,Sco]]},{})) &
  SState_ = SState &
  Out = no.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%                     PROOF OBLIGATIONS                        %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
% Besides proving that seccond and starprop are state invari-  %
% ants, we prove the following are also state invariants:      %
%                                                              %
% TypeInv1: Fo is a partial function                           %
%                                                              %
% TypeInv2: Fs is a partial function                           %
%                                                              %
% TypeInv3: the range of Br is a subset of the domain of M     %
%                                                              %
% TypeInv4: the range of Bw is a subset of the domain of M     %
%                                                              %
% TypeInv5: if [S,[O,L]] in Br then applyTo(Fo,O,L)            %
%                                                              %
% TypeInv6: if [S,[O,L]] in Bw then applyTo(Fo,O,L)            %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

typeInv5(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  foreach([S,[O,L]] in Br, applyTo(Fo,O,L)).

n_typeInv5(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(foreach([S,[O,L]] in Br, applyTo(Fo,O,L))).

typeInv6(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  foreach([S,[O,L]] in Bw, applyTo(Fo,O,L)).

n_typeInv6(SState) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  neg(foreach([S,[O,L]] in Bw, applyTo(Fo,O,L))).

% getRead

getReadPreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    getRead(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

getReadPreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    getRead(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

getReadPreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      getRead(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

getReadPreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      getRead(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

getReadPreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    getRead(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

getReadPreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    getRead(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

getReadPreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    getRead(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

getReadPreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    getRead(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

getReadCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(getReadPreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(getReadPreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(getReadPreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% getWrite

getWritePreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    getWrite(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

getWritePreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    getWrite(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

getWritePreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      getWrite(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

getWritePreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      getWrite(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

getWritePreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    getWrite(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

getWritePreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    getWrite(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

getWritePreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    getWrite(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

getWritePreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    getWrite(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

getWriteCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(getWritePreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(getWritePreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(getWritePreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.
  
% releaseRead

releaseReadPreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    releaseRead(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

releaseReadPreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    releaseRead(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

releaseReadPreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      releaseRead(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

releaseReadPreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      releaseRead(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

releaseReadPreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    releaseRead(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

releaseReadPreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    releaseRead(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

releaseReadPreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    releaseRead(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

releaseReadPreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    releaseRead(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

releaseReadCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(releaseReadPreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(releaseReadPreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(releaseReadPreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% releaseWrite

releaseWritePreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    releaseWrite(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

releaseWritePreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    releaseWrite(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

releaseWritePreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      releaseWrite(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

releaseWritePreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      releaseWrite(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

releaseWritePreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    releaseWrite(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

releaseWritePreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    releaseWrite(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

releaseWritePreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    releaseWrite(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

releaseWritePreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    releaseWrite(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

releaseWriteCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(releaseWritePreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(releaseWritePreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(releaseWritePreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% changeFo

changeFoPreservesTypeInv1(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    changeFo(SState,O,L,C,SState_,Out)
    implies  pfun(Fo_)
  ).

changeFoPreservesTypeInv2(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    changeFo(SState,O,L,C,SState_,Out)
    implies  pfun(Fs_)
  ).

changeFoPreservesTypeInv3(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      changeFo(SState,O,L,C,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

changeFoPreservesTypeInv4(SState,O,L,C,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      changeFo(SState,O,L,C,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

changeFoPreservesTypeInv5(SState,O,L,C,SState_,Out) :-
  neg(
    typeInv5(SState) &
    changeFo(SState,O,L,C,SState_,Out) 
    implies  typeInv5(SState_)
  ).

changeFoPreservesTypeInv6(SState,O,L,C,SState_,Out) :-
  neg(
    typeInv6(SState) &
    changeFo(SState,O,L,C,SState_,Out) 
    implies  typeInv6(SState_)
  ).

changeFoPreservesSeccond(SState,O,L,C,SState_,Out) :-
  neg(
    seccond(SState) &
    changeFo(SState,O,L,C,SState_,Out)
    implies  seccond(SState_)
  ).

changeFoPreservesStarprop(SState,O,L,C,SState_,Out) :-
  neg(
    starprop(SState) &
    changeFo(SState,O,L,C,SState_,Out)
    implies  starprop(SState_)
  ).

changeFoCheck(SState,O,L,C,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv1(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv2(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv3(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv4(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv5(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(changeFoPreservesTypeInv6(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(changeFoPreservesSeccond(SState,O,L,C,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(changeFoPreservesStarprop(SState,O,L,C,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% giveRW

giveRWPreservesTypeInv1(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    giveRW(SState,S,Si,O,X,SState_,Out)
    implies  pfun(Fo_)
  ).

giveRWPreservesTypeInv2(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    giveRW(SState,S,Si,O,X,SState_,Out)
    implies  pfun(Fs_)
  ).

giveRWPreservesTypeInv3(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      giveRW(SState,S,Si,O,X,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

giveRWPreservesTypeInv4(SState,S,Si,O,X,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      giveRW(SState,S,Si,O,X,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

giveRWPreservesTypeInv5(SState,S,Si,O,X,SState_,Out) :-
  neg(
    typeInv5(SState) &
    giveRW(SState,S,Si,O,X,SState_,Out) 
    implies  typeInv5(SState_)
  ).

giveRWPreservesTypeInv6(SState,S,Si,O,X,SState_,Out) :-
  neg(
    typeInv6(SState) &
    giveRW(SState,S,Si,O,X,SState_,Out) 
    implies  typeInv6(SState_)
  ).

giveRWPreservesSeccond(SState,S,Si,O,X,SState_,Out) :-
  neg(
    seccond(SState) &
    giveRW(SState,S,Si,O,X,SState_,Out)
    implies  seccond(SState_)
  ).

giveRWPreservesStarprop(SState,S,Si,O,X,SState_,Out) :-
  neg(
    starprop(SState) &
    giveRW(SState,S,Si,O,X,SState_,Out)
    implies  starprop(SState_)
  ).

giveRWCheck(SState,S,Si,O,X,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv1(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv2(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv3(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv4(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv5(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(giveRWPreservesTypeInv6(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(giveRWPreservesSeccond(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(giveRWPreservesStarprop(SState,S,Si,O,X,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% rescindRead

rescindReadPreservesTypeInv1(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    rescindRead(SState,S,Si,O,SState_,Out)
    implies  pfun(Fo_)
  ).

rescindReadPreservesTypeInv2(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    rescindRead(SState,S,Si,O,SState_,Out)
    implies  pfun(Fs_)
  ).

rescindReadPreservesTypeInv3(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      rescindRead(SState,S,Si,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

rescindReadPreservesTypeInv4(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      rescindRead(SState,S,Si,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

rescindReadPreservesTypeInv5(SState,S,Si,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    rescindRead(SState,S,Si,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

rescindReadPreservesTypeInv6(SState,S,Si,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    rescindRead(SState,S,Si,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

rescindReadPreservesSeccond(SState,S,Si,O,SState_,Out) :-
  neg(
    seccond(SState) &
    rescindRead(SState,S,Si,O,SState_,Out)
    implies  seccond(SState_)
  ).

rescindReadPreservesStarprop(SState,S,Si,O,SState_,Out) :-
  neg(
    starprop(SState) &
    rescindRead(SState,S,Si,O,SState_,Out)
    implies  starprop(SState_)
  ).

rescindReadCheck(SState,S,Si,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv1(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv2(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv3(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv4(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv5(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(rescindReadPreservesTypeInv6(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(rescindReadPreservesSeccond(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(rescindReadPreservesStarprop(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% rescindWrite

rescindWritePreservesTypeInv1(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    rescindWrite(SState,S,Si,O,SState_,Out)
    implies  pfun(Fo_)
  ).

rescindWritePreservesTypeInv2(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    rescindWrite(SState,S,Si,O,SState_,Out)
    implies  pfun(Fs_)
  ).

rescindWritePreservesTypeInv3(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      rescindWrite(SState,S,Si,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

rescindWritePreservesTypeInv4(SState,S,Si,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      rescindWrite(SState,S,Si,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

rescindWritePreservesTypeInv5(SState,S,Si,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    rescindWrite(SState,S,Si,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

rescindWritePreservesTypeInv6(SState,S,Si,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    rescindWrite(SState,S,Si,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

rescindWritePreservesSeccond(SState,S,Si,O,SState_,Out) :-
  neg(
    seccond(SState) &
    rescindWrite(SState,S,Si,O,SState_,Out)
    implies  seccond(SState_)
  ).

rescindWritePreservesStarprop(SState,S,Si,O,SState_,Out) :-
  neg(
    starprop(SState) &
    rescindWrite(SState,S,Si,O,SState_,Out)
    implies  starprop(SState_)
  ).

rescindWriteCheck(SState,S,Si,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv1(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv2(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv3(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv4(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv5(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(rescindWritePreservesTypeInv6(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(rescindWritePreservesSeccond(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(rescindWritePreservesStarprop(SState,S,Si,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% createObject

createObjectPreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    createObject(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

createObjectPreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    createObject(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

createObjectPreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      subset(DRBr,DM) &
      createObject(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

createObjectPreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      subset(DRBw,DM) &
      createObject(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

createObjectPreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    createObject(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

createObjectPreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    createObject(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

createObjectPreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    createObject(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

createObjectPreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    createObject(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

createObjectCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(createObjectPreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(createObjectPreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(createObjectPreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

% deleteObject

deleteObjectPreservesTypeInv1(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fo) &
    deleteObject(SState,S,O,SState_,Out)
    implies  pfun(Fo_)
  ).

deleteObjectPreservesTypeInv2(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    pfun(Fs) &
    deleteObject(SState,S,O,SState_,Out)
    implies  pfun(Fs_)
  ).

deleteObjectPreservesTypeInv3(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBr,DRBr,DM_,RBr_,DRBr_],
      dom(M,DM) & ran(Br,RBr) & dom(RBr,DRBr) &
      dom(M_,DM_) & ran(Br_,RBr_) & dom(RBr_,DRBr_),
      deleteObjectPreservesTypeInv5(SState,S,O,SState_,Out) &
      subset(DRBr,DM) &
      deleteObject(SState,S,O,SState_,Out)
      implies  subset(DRBr_,DM_)
    )
  ).

deleteObjectPreservesTypeInv4(SState,S,O,SState_,Out) :-
  SState = [[br,Br],[bw,Bw],[fo,Fo],[fs,Fs],[m,M]] &
  SState_ = [[br,Br_],[bw,Bw_],[fo,Fo_],[fs,Fs_],[m,M_]] &
  neg(
    let([DM,RBw,DRBw,DM_,RBw_,DRBw_],
      dom(M,DM) & ran(Bw,RBw) & dom(RBw,DRBw) &
      dom(M_,DM_) & ran(Bw_,RBw_) & dom(RBw_,DRBw_),
      deleteObjectPreservesTypeInv6(SState,S,O,SState_,Out) &
      subset(DRBw,DM) &
      deleteObject(SState,S,O,SState_,Out) 
      implies  subset(DRBw_,DM_)
    )
  ).

deleteObjectPreservesTypeInv5(SState,S,O,SState_,Out) :-
  neg(
    typeInv5(SState) &
    deleteObject(SState,S,O,SState_,Out) 
    implies  typeInv5(SState_)
  ).

deleteObjectPreservesTypeInv6(SState,S,O,SState_,Out) :-
  neg(
    typeInv6(SState) &
    deleteObject(SState,S,O,SState_,Out) 
    implies  typeInv6(SState_)
  ).

deleteObjectPreservesSeccond(SState,S,O,SState_,Out) :-
  neg(
    seccond(SState) &
    deleteObject(SState,S,O,SState_,Out)
    implies  seccond(SState_)
  ).

deleteObjectPreservesStarprop(SState,S,O,SState_,Out) :-
  neg(
    starprop(SState) &
    deleteObject(SState,S,O,SState_,Out)
    implies  starprop(SState_)
  ).

deleteObjectCheck(SState,S,O,SState_,Out) :-
  prolog_call(get_time(Ti)) &
  nl &
  write('Checking first type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv1(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking second type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv2(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking third type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv3(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fourth type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv4(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking fifth type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv5(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking sixth type invariant...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesTypeInv6(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking security condition...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesSeccond(SState,S,O,SState_,Out))
    & write('Done.') & nl &
  write('Checking *-property...') & prolog_call(flush_output)
    & naf(deleteObjectPreservesStarprop(SState,S,O,SState_,Out))
    & write('Done.') & nl & nl
  &  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) 
  &  write('Elapsed time: ') & write(T) & write(' seconds') & nl.

checkAll :-
  prolog_call(get_time(Ti)) &
  nl & write('Checking getRead...') & prolog_call(flush_output) &
  getReadCheck(SState,S,O,SState_,Out) &
  nl & write('Checking getWrite...') & prolog_call(flush_output) &
  getWriteCheck(SState,S,O,SState_,Out) &
  nl & write('Checking releaseRead...') & prolog_call(flush_output) &
  releaseReadCheck(SState,S,O,SState_,Out) &
  nl & write('Checking releaseWrite...') & prolog_call(flush_output) &
  releaseWriteCheck(SState,S,O,SState_,Out) &
  nl & write('Checking changeFoCheck...') & prolog_call(flush_output) &
  changeFoCheck(SState,O,L,C,SState_,Out) &
  nl & write('Checking giveRWCheck...') & prolog_call(flush_output) &
  giveRWCheck(SState,S,Si,O,X,SState_,Out) &
  nl & write('Checking rescindReadCheck...') & prolog_call(flush_output) &
  rescindReadCheck(SState,S,Si,O,SState_,Out) & 
  nl & write('Checking rescindWriteCheck...') & prolog_call(flush_output) &
  rescindWriteCheck(SState,S,Si,O,SState_,Out) &
  nl & write('Checking createObjectCheck...') & prolog_call(flush_output) &
  createObjectCheck(SState,S,O,SState_,Out) &
  nl & write('Checking deleteObjectCheck...') & prolog_call(flush_output) &
  deleteObjectCheck(SState,S,O,SState_,Out) &
  prolog_call(get_time(Te)) & prolog_call(T is Te - Ti) &
  nl & nl & write('All proof obligations discharged in ') & write(T) & 
  write(' seconds') & nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                              %
%                    AUXILIARY PREDICATES                      %
%                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

n_applyTo(F,X,Y) :- ncomp({[X,X]},F,{[X,Y]}).


